{-
 -
 - APCC
 - (C) Ivana Kojovic, DTU 2012
 - All Rights Reserved.
 
 - 22nd April -  remove type check condition in unifyAux
 -              
 - 
 -}

-- | Module @EncryptedPattern@ represents theorem 2 checker: disjointness from encrytptions, EMPi sets and EST for single protocol
module EncryptedPattern (makeEST,unifyEMP, unionOfMaps, State', StateEMP ) where
import Message
import FormatTypeSafe
import Unification
import qualified Data.Map as Map
import Control.Monad.State
import Data.List
import Debug.Trace

type State' = (Map.Map (Message,String) MessageType,Int)
type StateEMP = ([Map.Map (Message,String) MessageType], Int)

-- | Given a key and a message, function makes new message symmetrically encrypted message
encryptMsg :: ((Message,String), MessageType) -> String -> ((Message,String), MessageType)
encryptMsg ((msg,protocolName),msgType) newKey = ((Symcrypt (Atom (Ident newKey)) msg, protocolName), Scrypt (Basic Symkey) msgType)


-- | Extension of the theorem. Number of EMPs is equal to the depth of encryptions inside a protocol.
depth :: ((Message,String),MessageType) -> Int
depth ((Atom _, protocolName), _)= 0
depth ((Hash id t,protocolName), HashT tt) = depth ((t,protocolName),tt)
depth ((Concat t,protocolName), Conc tt) = (foldl (\acc (a,a1) -> depth (a,a1)) 0 (zip (zipName t protocolName) tt))
depth ((Asycrypt k t,protocolName), Crypt kt tt) = depth ((k,protocolName),kt) + depth ((t,protocolName),tt)
depth ((Symcrypt k t,protocolName), Scrypt kt tt) = (depth ((k,protocolName),kt) + depth ((t,protocolName),tt))+1

-- | Uses depth function

depthProtocol :: [((Message,String), MessageType)] -> Int  
depthProtocol array = maximum (map depth array)

-- | It gives single EMPi, preserving counter value for alpha renaming

makeSingleEMP :: [((Message,String),MessageType)] -> State State' State'
makeSingleEMP [] = do
   (state,counter) <- get
   return (state,counter)

makeSingleEMP (x:xs) = do
     (state, counter) <- get
     let aux = encryptMsg x ("KEY"++(show counter))
      in  put(Map.insert ((alphaRenaming (fst (fst aux)) counter), (snd(fst aux))) (snd aux) $ state ,counter+1)
     makeSingleEMP xs

-- | Now we make EMP set. Monadic computation.
--

computeEMP :: Int -> State StateEMP StateEMP

computeEMP 0 = do
            (state, counter) <- get
            return (state, counter)

computeEMP depth = do
                (state, counter) <- get
                let newState = evalState (makeSingleEMP (Map.toList(head(state)))) (Map.empty, counter)
                 in put((fst newState):state, (snd newState)+counter)
                computeEMP (depth-1)


-- Check whether exist unification between elements EMPi. Function calls UnifyEMPij and unifyAux

unifyEMP :: [Map.Map (Message,String) MessageType] -> Bool
unifyEMP [] = False
unifyEMP (x: arraytail) = (foldl (\acc a -> if (unifyEMPij x a) then True else acc ) False arraytail) || unifyEMP arraytail


unifyEMPij :: Map.Map (Message,String) MessageType -> Map.Map (Message,String) MessageType -> Bool
unifyEMPij x y = if ((not (Map.null x)) && (not (Map.null y)))
                                  then let new = head(Map.toList x)
                                        in (unifyAux new y (Map.keys y)) || (unifyEMPij (Map.delete (fst new) x) y)
                                  else False




unifyAux :: ((Message,String),MessageType) -> Map.Map (Message,String) MessageType -> [(Message,String)] -> Bool
unifyAux (msg,msgType) table array =  foldl (\acc a -> if ((unify[(fst msg,fst a)] /= Nothing) && (msgType == table Map.!a)) -- msg is message+protocolname!!
                                                            then error ("\n Disjointness is violated!  \n Unification between terms: "++ show (fst msg) ++" in protocol "++ show (snd msg) ++ " and " ++ show (fst a) ++ " in protocol " ++ show (snd a)) 
                                                            else acc) False array
  
-- | Filter for Map.map data type. It filters all the non atomic subterms
isNonAtomSubTerm:: (Message,String) -> Bool
isNonAtomSubTerm ((Atom (Mapping _ _)),protocolName) = False
isNonAtomSubTerm ((Atom (Ident t)),protocolName) = False
isNonAtomSubTerm _ = True

-- | Helpful function to make one single Map.map data type from array of Map.maps by merging them.
unionOfMaps :: ([Map.Map (Message,String) MessageType],num) -> (Map.Map (Message,String) MessageType, num)
unionOfMaps ([],num) = (Map.empty,num)
unionOfMaps ((x:xs),num) = (Map.union x (fst(unionOfMaps (xs,num))),num)


-- | Forming a EST set, keeping counter still. Making EST doesnt have sense if we can unify EMPi elements..
makeEST :: ([((Message,String), MessageType)],Int) -> (Map.Map (Message,String) MessageType, Int) --- includes unify checking for the EMPi-s
makeEST (array,counter) =  if theoremNum1 array
                              then let aux = protocolMP array (Map.empty, counter)
                                       aux2 = (evalState (computeEMP $ depthProtocol array) ([fst aux], snd aux ))
                                       aux3 = (protocolNVST array (Map.empty, snd aux2))
                                       aux3filter = Map.filterWithKey (\k _ -> isNonAtomSubTerm k) (fst aux3)
                                     in if (unifyEMP (fst aux2))
                                               then error "Disjointness is violated!" -- never get here!!
                                               else unionOfMaps ((aux3filter):(init (fst(aux2))), snd aux3)
                              else error "Some of the protocols is not format type safe" -- never get here!
                             

--------------------- helpful function
--
putNameInsideMessages :: ([(Message,MessageType)], String) -> [((Message,String),MessageType)]
putNameInsideMessages ([],protocolName) = []
putNameInsideMessages ((msg,msgType):arrayTail, protocolName) = ((msg,protocolName),msgType): (putNameInsideMessages (arrayTail,protocolName))

----------- tests for module ---------------------------------------------------------------------------------------------------------------------------------------------------

