{-
 -
 - APCC
 - (C) Ivana Kojovic, DTU 2012
 - All Rights Reserved.
 -
 - Unification by the book "Message rewriting and all that"
 -
13th april --  added part for mapping in occursCheck
               added part for mapping in the substitution
               added part for mapping in the unify' part + condition next to isConstant

27th april --- added mapping as the function + occursCheck part for it.. 
28th april --- added improvments for mapping and occursCheck, substitution improved
               
-}

-- | Module @Unification@ that operates on message terms. Algorithm is modelled according to book "Message rewriting and all that"

module Unification (unify) where

import Message


-- | Substitution of the messages on single term
substitution :: Message -> Message -> Message -> Message
substitution (Atom (Ident x)) t (Atom (Ident s))  | s == x    = t      
                                                  | otherwise = Atom (Ident s)

-- | Substitution of the messages on complex (function) term
substitution x t (Atom (Mapping id a)) = Atom (Mapping id (map toAtomic (map (substitution x t) (wrapIdent a))))
substitution x t (Concat xx)       = Concat (map (substitution x t) xx) 
substitution x t (Asycrypt k xx)   = Asycrypt  (substitution x t k) (substitution x t xx)
substitution x t (Symcrypt k xx)   = Symcrypt  (substitution x t k) (substitution x t xx)
substitution x t (Hash _ xx)         = substitution x t xx

substitutionp x t (p1,p2)    = (substitution x t p1, substitution x t p2)



-- | if variable is "inside" the non atomic term on the right side -}

occursCheck:: (String, Message) -> Bool

occursCheck (x, Atom(Ident t)) | x==t = True 
                               | otherwise = False

occursCheck (x, Atom (Mapping k t)) = let t2 = wrapIdent t
                                       in foldl (\acc a -> if occursCheck (x,a) then True else acc) False t2
occursCheck (x, Concat t) = foldl (\acc a -> if occursCheck (x,a) then True else acc) False t

occursCheck (x, Asycrypt k t)  = occursCheck (x,t) || occursCheck (x,k)
occursCheck (x, Symcrypt k t)  = occursCheck (x,t) || occursCheck (x,k)
occursCheck (x, Hash _ t)      = occursCheck (x,t)
---------------------------------   Wrap message  ---------technical stuff, not so important

wrapIdent :: [Atomic] -> [Message]
wrapIdent [] = []
wrapIdent (x:xs) = (Atom x): wrapIdent xs

 --------------------------------- Making atomic message ---- also not so important

toAtomic :: Message -> Atomic
toAtomic (Atom (Ident t)) = Ident t
toAtomic (Atom (Mapping id t)) = Mapping id t


-- | Unification algorithm also works for set of the messages

unify :: [(Message, Message)] -> Maybe [(Message, Message)]

unify xx = unify' xx []

unify' :: [(Message, Message)] -> [(Message, Message)] -> Maybe [(Message, Message)]

-- | When there is nothing to process, acc is the solution
unify' [] acc = Just acc

-- | Delete rule
unify' ((a, b):xx) acc | a==b = unify' xx acc


-- | Substitute rule : takes care of occurrence of the variable of on the right side; constants on the left side => do the substitution with switched arguments

unify' ((Atom(Ident x),t):xx) acc | not (occursCheck (x,t)) && isVariable x = -- SUBSTITUTE - watching out on constants!
                                    unify' (map (substitutionp (Atom (Ident x)) t) xx) ((Atom(Ident x),t) : map (substitutionp (Atom (Ident x)) t) acc)
                                  | isVarTerm t && not (isVariable x)  =  
                                          unify' (map (substitutionp t (Atom (Ident x))) xx) ((t,(Atom (Ident x))) : map (substitutionp t  (Atom (Ident x))) acc)

                                  | otherwise = Nothing


-- | Orient rule
unify' ((t,Atom(Ident x)):xx) acc = unify' ((Atom(Ident x),t):xx) acc 


-- | Decompose rules for all the non atomic message terms, including mappings!
unify' ((Asycrypt key1 mess1, Asycrypt key2 mess2):xx) acc 
      = unify' ([(key1,key2)]++([(mess1,mess2)] ++ xx)) acc 

unify' ((Symcrypt key1 mess1, Symcrypt key2 mess2):xx) acc  
      = unify' ([(key1,key2)]++([(mess1,mess2)] ++ xx)) acc 
    

unify' ((Hash id1 mess1, Hash id2 mess2):xx) acc 
    | id1 == id2 = unify' ([(mess1,mess2)] ++ xx) acc
    | otherwise = Nothing

unify' ((Concat f1s, Concat f2s):xx) acc 
    | length f1s == length f2s = unify' (zip f1s f2s ++ xx) acc 
    | otherwise = Nothing


unify' ((Atom (Mapping id1 t1), Atom (Mapping id2 t2)):xx) acc
    | (id1 == id2 && length t1 == length t2) = unify' (zip (wrapIdent t1) (wrapIdent t2) ++ xx) acc
    | otherwise = Nothing


-- | For everything else, return Nothing
unify' (_) acc = Nothing  

{- ----------------------------------------------- test-------------------------------------------------------------- -}








